@@ -30,6 +30,7 @@
 
 #undef GTK_DISABLE_DEPRECATED
 
+#include <config.h>
 #include <string.h>
 
 #include "gtkarrow.h"
@@ -45,9 +46,11 @@
 #include "gdk/gdkkeysyms.h"
 #include "gtkcombo.h"
 #include "gtkframe.h"
+#include "gtkprivate.h"
 #include "gtkintl.h"
+#include "gtkalias.h"
 
-const gchar *gtk_combo_string_key = "gtk-combo-string-value";
+const gchar gtk_combo_string_key[] = "gtk-combo-string-value";
 
 #define COMBO_LIST_MAX_HEIGHT	(400)
 #define	EMPTY_LIST_HEIGHT	(15)
@@ -139,41 +142,41 @@ gtk_combo_class_init (GtkComboClass * klass)
 
   g_object_class_install_property (gobject_class,
                                    PROP_ENABLE_ARROW_KEYS,
-                                   g_param_spec_boolean ("enable_arrow_keys",
+                                   g_param_spec_boolean ("enable-arrow-keys",
                                                          P_("Enable arrow keys"),
                                                          P_("Whether the arrow keys move through the list of items"),
                                                          TRUE,
-                                                         G_PARAM_READABLE | G_PARAM_WRITABLE));
+                                                         GTK_PARAM_READWRITE));
   g_object_class_install_property (gobject_class,
                                    PROP_ENABLE_ARROWS_ALWAYS,
-                                   g_param_spec_boolean ("enable_arrows_always",
+                                   g_param_spec_boolean ("enable-arrows-always",
                                                          P_("Always enable arrows"),
                                                          P_("Obsolete property, ignored"),
                                                          TRUE,
-                                                         G_PARAM_READABLE | G_PARAM_WRITABLE));
+                                                         GTK_PARAM_READWRITE));
   g_object_class_install_property (gobject_class,
                                    PROP_CASE_SENSITIVE,
-                                   g_param_spec_boolean ("case_sensitive",
+                                   g_param_spec_boolean ("case-sensitive",
                                                          P_("Case sensitive"),
                                                          P_("Whether list item matching is case sensitive"),
                                                          FALSE,
-                                                         G_PARAM_READABLE | G_PARAM_WRITABLE));
+                                                         GTK_PARAM_READWRITE));
 
   g_object_class_install_property (gobject_class,
                                    PROP_ALLOW_EMPTY,
-                                   g_param_spec_boolean ("allow_empty",
+                                   g_param_spec_boolean ("allow-empty",
                                                          P_("Allow empty"),
 							 P_("Whether an empty value may be entered in this field"),
                                                          TRUE,
-                                                         G_PARAM_READABLE | G_PARAM_WRITABLE));
+                                                         GTK_PARAM_READWRITE));
 
   g_object_class_install_property (gobject_class,
                                    PROP_VALUE_IN_LIST,
-                                   g_param_spec_boolean ("value_in_list",
+                                   g_param_spec_boolean ("value-in-list",
                                                          P_("Value in list"),
                                                          P_("Whether entered values must already be present in the list"),
                                                          FALSE,
-                                                         G_PARAM_READABLE | G_PARAM_WRITABLE));
+                                                         GTK_PARAM_READWRITE));
   
    
   oclass->destroy = gtk_combo_destroy;
@@ -513,8 +516,9 @@ gtk_combo_get_pos (GtkCombo * combo, gint * x, gint * y, gint * height, gint * w
 }
 
 static void
-gtk_combo_popup_list (GtkCombo * combo)
+gtk_combo_popup_list (GtkCombo *combo)
 {
+  GtkWidget *toplevel;
   GtkList *list;
   gint height, width, x, y;
   gint old_width, old_height;
@@ -553,6 +557,15 @@ gtk_combo_popup_list (GtkCombo * combo)
     }
   
   gtk_window_move (GTK_WINDOW (combo->popwin), x, y);
+
+  toplevel = gtk_widget_get_toplevel (GTK_WIDGET (combo));
+
+  if (GTK_IS_WINDOW (toplevel))
+    {
+      gtk_window_group_add_window (_gtk_window_get_group (GTK_WINDOW (toplevel)), 
+                                   GTK_WINDOW (combo->popwin));
+    }
+
   gtk_widget_set_size_request (combo->popwin, width, height);
   gtk_widget_show (combo->popwin);
 
@@ -580,6 +593,8 @@ gtk_combo_popdown_list (GtkCombo *combo)
     }
   
   gtk_widget_hide (combo->popwin);
+
+  gtk_window_group_add_window (_gtk_window_get_group (NULL), GTK_WINDOW (combo->popwin));
 }
 
 static gboolean
@@ -1240,3 +1255,6 @@ gtk_combo_get_property (GObject         *object,
     }
    
 }
+
+#define __GTK_SMART_COMBO_C__
+#include "gtkaliasdef.c"
